/* date_offset - display the date and time, with an offset
**
** Copyright (C) 1998,2025 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** You may use this software however you like as long as you keep my
** name on it and don't sue me.
**
** For commentary on this license please see http://acme.com/license.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdbool.h>


#define MOD(x,m) ( ( (x) % (m) + (m) ) % (m) )	/* handle negatives right */

static char* argv0;


static void usage( void )
    {
    (void) fprintf( stderr, "usage:  %s [-u] [-f format] offset\n", argv0 );
    exit( 1 );
    }


static bool is_leap_year( int year )
    {
    return year % 400 ? ( year % 100 ? ( year % 4 ? false : true ) : false ) : true;
    }


static int days_in_month( int year, int month )
    {
    int month_days[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
    if ( month == 2 && is_leap_year( year ) )
	return 29;
    else
	return month_days[month];
    }


int main( int argc, char** argv )
    {
    int argn;
    int gmt;
    char* format;
    char* offset;
    int offset_sign = 1;
    int offset_val;
    char offset_type;
    int offset_mult = 1;
    time_t t;
    struct tm* tmP;
    int new_mon, days;
    char buf[1000];

    argv0 = argv[0];
    argn = 1;
    gmt = 0;
    /* Default format is Wdy Mon  d hh:mm:ss Zon Year. */
    format = "%a %b %e %H:%M:%S %Z %Y";
    while ( argn < argc && argv[argn][0] == '-' && argv[argn][1] >= 'a' && argv[argn][1] <= 'z' )
	{
	if ( strncmp( argv[argn], "-u", 2 ) == 0 )
	    gmt = 1;
	else if ( strncmp( argv[argn], "-f", 2 ) == 0 )
	    {
	    ++argn;
	    if ( argn >= argc )
		usage();
	    format = argv[argn];
	    }
	else
	    usage();
	++argn;
	}
    if ( argn != argc - 1 )
	usage();

    offset = argv[argn];
    switch ( offset[0] )
	{
	case '+': offset_sign = 1; break;
	case '-': offset_sign = -1; break;
	default: usage();
	}
    offset_type = tolower( offset[ strlen( offset ) - 1 ] );
    switch ( offset_type )
	{
	case 's': offset_mult = 1; break;
	case 'm': offset_mult = 60; break;
	case 'h': offset_mult = 3600; break;
	case 'd': offset_mult = 86400; break;
	case 'w': offset_mult = 86400 * 7; break;
	case 'o': case 'y': break;
	default: usage();
	}
    offset_val = atoi( &offset[1] );

    t = time( (time_t*) 0 );
    if ( offset_type != 'o' && offset_type != 'y' )
	t += offset_val * offset_mult * offset_sign;

    if ( gmt )
	tmP = gmtime( &t );
    else
	tmP = localtime( &t );

    if ( offset_type == 'o' )
	{
	tmP->tm_mon += offset_val * offset_sign;
	if ( tmP->tm_mon < 0 || tmP->tm_mon > 11 )
	    {
	    /* Fix up month and year. */
	    new_mon = MOD( tmP->tm_mon, 12 );
	    tmP->tm_year += ( tmP->tm_mon - new_mon ) / 12;
	    tmP->tm_mon = new_mon;
	    }
	}
    else if ( offset_type == 'y' )
	tmP->tm_year += offset_val * offset_sign;

    /* Correct for days of the month which don't exist. */
    days = days_in_month( tmP->tm_year + 1900, tmP->tm_mon + 1 );
    if ( tmP->tm_mday > days )
	tmP->tm_mday = days;

    (void) strftime( buf, sizeof(buf), format, tmP );
    (void) printf( "%s\n", buf );
    exit( 0 );
    }
